import type { HanziData } from '../../../src-common/hanzi-data'
import { lookup } from './lut'
import { readFile } from 'fs/promises'

export interface GetHanziDataOpts {
  char: string
  kind?: GetHanziDataKind
}

export const enum GetHanziDataKind {
  ANY = 'any',
  DEFAULT = 'default',
  TRADITIONAL = 'traditional'
}

export async function getHanziProp (
  { prop, ...opts }: GetHanziDataOpts & { prop: string }
): Promise<string | string[] | undefined> {
  const data = await getHanziData(opts)
  if (data === undefined) {
    return undefined
  }
  return data[prop] ?? lookup(data, prop)
}

const dataFilePrefix = '_hd'
const jsonpCallbackFnName = 'hd'
const cached: Record<string, HanziData> = {}
const canAccessFilesystem = typeof document === 'undefined'

export async function getHanziData (
  { char, kind }: GetHanziDataOpts
): Promise<HanziData> {
  if (char.length !== 1) {
    throw new Error(`Can only get data for single chars, got: ${char}`)
  }

  const url = process.env.NODE_ENV === 'production'
    // actual Anki keeps only the last path component
    ? `${dataFilePrefix}${char}.js`
    : `/generated/hanzi-data/${dataFilePrefix}${char}.js`
  let data: HanziData
  if (canAccessFilesystem) {
    // read directly from disk when running in test
    const relative = url.substring(1)
    const jsonp = await readFile(relative, { encoding: 'utf-8' })
    data = JSON.parse(jsonp.substring(
      'hd("A",'.length,
      jsonp.length - ')'.length
    ))
  } else {
    if (char in cached) {
      data = { ...cached[char] }
    } else {
      // in production and parcel use a JSONP-like API because everything else
      // failed in AnkiDroid
      data = await fetchJsonp(url, char)
      cached[char] = { ...data }
    }
  }
  if (kind === GetHanziDataKind.TRADITIONAL && data.trad !== undefined) {
    for (const overridden in data.trad) {
      data[overridden] = data.trad[overridden]
    }
  }
  return data
}

/** chars against functions to call on resolve. */
const jsonpWaiters = new Map<string, Array<(HanziData) => void>>()

/**
 * Called from the JSONP script with the data.
 *
 * Raw JSON would not work with AnkiDroid, which does not support XHR/fetch,
 * but allows scripts.
 */
if (!canAccessFilesystem) {
  if (typeof window[jsonpCallbackFnName] !== 'undefined') {
    throw new Error(
      `Cannot register jsonp callback because window.${
        jsonpCallbackFnName
      } is used`
    )
  }
  window[jsonpCallbackFnName] = (char: string, data: HanziData) => {
    (jsonpWaiters.get(char) ?? []).forEach(w => { w(data) })
    jsonpWaiters.delete(char)
  }
}

/**
 * Registers a waiter for the specified char.
 *
 * It will be removed automatically after the data has been received.
 */
function registerJsonpWaiter (
  char: string,
  callback: (data: HanziData) => void
): void {
  jsonpWaiters.set(
    char,
    [...(jsonpWaiters.get(char) ?? []), callback]
  )
}

async function fetchJsonp (url: string, char: string): Promise<HanziData> {
  return await new Promise((resolve, reject) => {
    if (document.querySelector(`script[src="${url}"]`) !== null) {
      // already being fetched, append to the list of existing waiters
      registerJsonpWaiter(char, resolve)
    } else {
      // no fetch in flight, inject a new script tag
      const script = document.createElement('script')
      registerJsonpWaiter(
        char,
        data => {
          script.remove()
          resolve(data)
        }
      )
      script.async = true
      script.src = url
      script.onerror = () => {
        script.remove()
        reject(new Error(`No character data available for ${char}`))
      }
      document.body.appendChild(script)
    }
  })
}
